﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.BinaryVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Knapsack {
  /// <summary>
  /// An operator that improves knapsack solutions.
  /// </summary>
  [Item("KnapsackImprovementOperator", "An operator that improves knapsack solutions.")]
  [StorableClass]
  public sealed class KnapsackImprovementOperator : SingleSuccessorOperator, IImprovementOperator {
    #region Parameter properties
    public ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public ILookupParameter<IntValue> KnapsackCapacityParameter {
      get { return (ILookupParameter<IntValue>)Parameters["KnapsackCapacity"]; }
    }
    public ILookupParameter<DoubleValue> PenaltyParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Penalty"]; }
    }
    public IValueLookupParameter<IItem> TargetParameter {
      get { return (IValueLookupParameter<IItem>)Parameters["Target"]; }
    }
    public ILookupParameter<IntArray> ValuesParameter {
      get { return (ILookupParameter<IntArray>)Parameters["Values"]; }
    }
    public ILookupParameter<IntArray> WeightsParameter {
      get { return (ILookupParameter<IntArray>)Parameters["Weights"]; }
    }
    #endregion

    #region Properties
    private IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }
    private IntValue KnapsackCapacity {
      get { return KnapsackCapacityParameter.ActualValue; }
      set { KnapsackCapacityParameter.ActualValue = value; }
    }
    private DoubleValue Penalty {
      get { return PenaltyParameter.ActualValue; }
      set { PenaltyParameter.ActualValue = value; }
    }
    private IItem Target {
      get { return TargetParameter.ActualValue; }
    }
    private IntArray Values {
      get { return ValuesParameter.ActualValue; }
      set { ValuesParameter.ActualValue = value; }
    }
    private IntArray Weights {
      get { return WeightsParameter.ActualValue; }
      set { WeightsParameter.ActualValue = value; }
    }
    #endregion

    [StorableConstructor]
    private KnapsackImprovementOperator(bool deserializing) : base(deserializing) { }
    private KnapsackImprovementOperator(KnapsackImprovementOperator original, Cloner cloner) : base(original, cloner) { }
    public KnapsackImprovementOperator()
      : base() {
      #region Create parameters
      Parameters.Add(new ScopeParameter("CurrentScope"));
      Parameters.Add(new LookupParameter<IntValue>("KnapsackCapacity"));
      Parameters.Add(new LookupParameter<DoubleValue>("Penalty"));
      Parameters.Add(new ValueLookupParameter<IItem>("Target"));
      Parameters.Add(new LookupParameter<IntArray>("Values"));
      Parameters.Add(new LookupParameter<IntArray>("Weights"));
      #endregion
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new KnapsackImprovementOperator(this, cloner);
    }

    public override IOperation Apply() {
      BinaryVector sol = CurrentScope.Variables[TargetParameter.ActualName].Value as BinaryVector;

      // calculate value-to-weight ratio
      double[] ratio = new double[Values.Length];
      for (int i = 0; i < ratio.Length; i++)
        ratio[i] = (double)Values[i] / (double)Weights[i];


      // calculate order for ratio
      int[] order = new int[ratio.Length];
      ratio.Select((x, index) => new { Value = x, ValueIndex = index })
           .OrderBy(x => x.Value)
           .Select((x, index) => new { ValueIndex = x.ValueIndex, ItemIndex = index }).ToList()
           .ForEach(x => order[x.ItemIndex] = x.ValueIndex);

      int j = sol.Length - 1;
      while (KnapsackEvaluator.Apply(sol, KnapsackCapacity, Penalty, Weights, Values)
                              .SumWeights.Value > KnapsackCapacity.Value && j >= 0)
        sol[order[j--]] = false;

      // calculate weight
      int weight = 0;
      for (int i = 0; i < sol.Length; i++)
        if (sol[i]) weight += Weights[i];

      // improve solution
      bool feasible = true; j = 0;
      while (feasible && j < sol.Length) {
        while (sol[order[j]]) j++;
        if (weight + Weights[order[j]] <= KnapsackCapacity.Value) {
          sol[order[j]] = true;
          weight += Weights[order[j]];
        } else feasible = false;
      }

      CurrentScope.Variables[TargetParameter.ActualName].Value = sol;

      return base.Apply();
    }
  }
}
